<!DOCTYPE html>
<html lang="en">
    <head>
      <meta charset="utf-8">
      {{ bokeh_css }}
      {{ bokeh_js }}
      <style type="text/css">{% include 'styles.css' %}</style>
    </head>
    <body>
      <div class="container" style="width: 80%; margin: 0 auto 1.5rem auto !important">
        {{ plot_div|indent(8) }}
        {{ plot_script|indent(8) }}
      </div>
    </body>
</html>
